; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.

; I/O primitives

; See doc/io.txt for a description of Scheme 48's I/O system.

; Argument specs

(define open-input-port->
  (input-type (lambda (x)
		(and (port? x)
		     (port-has-status? x (enum port-status-options
					       open-for-input))))
	      no-coercion))

(define open-output-port->
  (input-type (lambda (x)
		(and (port? x)
		     (port-has-status? x (enum port-status-options
					       open-for-output))))
	      no-coercion))

(define channel->     (input-type channel?        no-coercion))

(define (port-has-status? port status)
  (not (= 0 (bitwise-and (extract-fixnum (port-status port))
			 (ashl 1 status)))))

; Check SPEC type and then call OPEN-CHANNEL.

(define-consing-primitive open-channel (any-> fixnum->)
  (lambda (ignore) channel-size)
  (lambda (spec mode key)
    (if (cond ((or (= mode (enum open-channel-option input-file))
		   (= mode (enum open-channel-option output-file)))
	       (vm-string? spec))
	      ((or (= mode (enum open-channel-option raw-input-channel))
		   (= mode (enum open-channel-option raw-output-channel))
		   (= mode (enum open-channel-option special-input-channel))
		   (= mode (enum open-channel-option special-output-channel)))
	       (fixnum? spec))  ; works for Unix, what to do in general?
	      (else #f))
	(open-channel spec mode key
		      collect-saving-temp      ; a GC may be needed
		      (lambda (channel)        ; all is okay
			(goto return channel))
		      (lambda (spec reason)    ; something went wrong
			(if (= reason (enum exception cannot-open-channel))
			    (goto return false)
			    (raise-exception* reason 0 spec (enter-fixnum mode)))))
	(raise-exception wrong-type-argument 0 spec (enter-fixnum mode)))))

(define-primitive close-channel (channel->)
  (lambda (channel)
    (if (close-channel channel)
	(return unspecific)
	(raise-exception os-error 0 channel))))

(define (channel-read-or-write read?)
  (lambda (thing start count channel)
    (let* ((lose (lambda (reason)
		   (raise-exception* reason 0
				     thing
				     (enter-fixnum start)
				     (enter-fixnum count)
				     channel))))
      (if (and (if read?
		   (and (input-channel? channel)
			(not (immutable? thing)))
		   (output-channel? channel))
	       (open? channel)
	       (or (vm-string? thing)
		   (code-vector? thing)))
	  (let ((length (if (vm-string? thing)
			    (vm-string-length thing)
			    (code-vector-length thing)))
		(addr (+ (address-after-header thing) start))
		(channel (extract-channel channel)))
	    (if (< length (+ start count))
		(lose (enum exception index-out-of-range))
		((if read? do-channel-read do-channel-write)
		 channel addr count lose)))
	  (lose (enum exception wrong-type-argument))))))

(define (do-channel-read channel start count lose)
  (read-block channel start count
	      (lambda (okay? eof? got)
		(cond ((not okay?)
		       (lose (enum exception os-error)))
		      (eof?
		       (goto return eof-object))
		      (else
		       (goto return-fixnum got))))))

(define (do-channel-write channel start count lose)
  (write-block channel start count
	       (lambda (okay? got)
		 (if okay?
		     (goto return-fixnum got)
		     (lose (enum exception os-error))))))

(let ((do-it (channel-read-or-write #t)))
  (define-primitive channel-read (any-> fixnum-> fixnum-> channel->) do-it))
  
(let ((do-it (channel-read-or-write #f)))	   
  (define-primitive channel-write (any-> fixnum-> fixnum-> channel->) do-it))

(define-consing-primitive find-ready-channels (any-> any-> any->)
  (lambda (channels) (* (vm-length channels) vm-pair-size))
  (lambda (time-mantissa time-exponent channels key)
    (let ((do-it (lambda (time)
		   (let ((ready (find-ready-channels channels time key)))
		     (if (false? ready)
			 (raise-exception os-error 0
					  time-mantissa
					  time-exponent
					  channels)
			 (goto return ready))))))
      (cond ((false? time-mantissa)
	     (do-it -1))
	    ((and (fixnum? time-mantissa)
		  (fixnum? time-exponent))
	     (do-it (adjust-time (extract-fixnum time-mantissa)
				 (extract-fixnum time-exponent))))
	    (else
	     (raise-exception wrong-type-argument 0
			      time-mantissa time-exponent channels))))))

(define (vm-length list)
  (do ((l list (vm-cdr l))
       (len 0 (+ len 1)))
      ((vm-eq? null l) len)))

;----------------------------------------------------------------
; Port instructions.

(define (read-or-peek-char read?)
  (lambda (port)
    (let ((i (extract-fixnum (port-in-index port)))
	  (l (extract-fixnum (port-in-limit port)))
	  (b (port-in-buffer port)))
      (cond ((not (= i l))
	     (if read?
		 (set-port-in-index! port (enter-fixnum (+ i 1))))
	     (goto return-char (ascii->char (code-vector-ref b i))))
	    (else
	     (raise-exception buffer-full/empty 0 port))))))

(let ((do-it (read-or-peek-char #t)))
  (define-primitive read-char (open-input-port->) do-it))

(let ((do-it (read-or-peek-char #f)))
  (define-primitive peek-char (open-input-port->) do-it))

(define-primitive char-ready? (open-input-port->)
  (lambda (port)
    (if (= (extract-fixnum (port-in-index port))
	   (extract-fixnum (port-in-limit port)))
	(raise-exception buffer-full/empty 0 port)
	(goto return-boolean #t))))
    
(define-primitive write-char (char-> open-output-port->)
  (lambda (char port)
    (let ((i (extract-fixnum (port-out-index port)))
	  (b (port-out-buffer port)))
      (cond ((not (= i (code-vector-length b)))
	     (set-port-out-index! port (enter-fixnum (+ i 1)))
	     (code-vector-set! (port-out-buffer port) i (char->ascii char))
	     (goto return-unspecific 0))
	    (else
	     (raise-exception buffer-full/empty 0 (enter-char char) port))))))

; Reading and writing a bunch of characters.
; ({read|write}-chars {string|code-vector} start count port)

(define (read-or-write-chars do-read-or-write)
  (lambda (thing start count port)
    (let* ((lose (lambda (reason to-do)
		   (raise-exception* reason 0
				     thing
				     (enter-fixnum (+ start (- count to-do)))
				     (enter-fixnum count)
				     port)))
	   (do-it (lambda (length)
		    (if (< length (+ start count))
			(lose (enum exception index-out-of-range) count)
			(do-read-or-write
			 (+ (address-after-header thing) start)
			 count
			 port
			 lose)))))
      (cond ((vm-string? thing)
	     (do-it (vm-string-length thing)))
	    ((code-vector? thing)
	     (do-it (code-vector-length thing)))
	    (else
	     (lose (enum exception wrong-type-argument) count))))))
  
(define (do-read-chars start count port lose)
  (let* ((i (extract-fixnum (port-in-index port)))
	 (l (extract-fixnum (port-in-limit port)))
	 (limit (min count (- l i))))
    (copy-memory! (addr+ (port-in-buffer port) i) start limit)
    (set-port-in-index! port (enter-fixnum (+ i limit)))
    (if (= limit count)
	(goto return-fixnum 0)
	(lose (enum exception buffer-full/empty) (- count limit)))))

(define (do-write-chars start count port lose)
  (let ((b (port-out-buffer port))
	(i (extract-fixnum (port-out-index port))))
    (cond ((> (code-vector-length b) (+ i count))
	   (copy-memory! start (addr+ (address-after-header b) i) count)
	   (set-port-out-index! port (enter-fixnum (+ i count)))
	   (goto return-unspecific 0))
	  (else
	   (lose (enum exception buffer-full/empty) count)))))

(let ((do-it (read-or-write-chars do-read-chars)))
  (define-primitive xread-block (any-> fixnum-> fixnum-> open-input-port->)
    do-it))
	   
(let ((do-it (read-or-write-chars do-write-chars)))
  (define-primitive write-block (any-> fixnum-> fixnum-> open-output-port->)
    do-it))
	   
; *** our entry for the obscure comment of the year contest.
;
; pclsring is the term in its for the mechanism that makes the operating
; appear to be a virtual machine.  the paradigm is that of the blt instruction
; on the pdp-10: its arguments are in a set of registers, and if the instruction
; gets interrupted in the middle, the registers reflect the intermediate state;
; the pc is set to the blt instruction itself, and the process can be resumed
; in the usual way.
; for more on pclsring see `pclsring: keeping process state modular' by alan
; bawden (ftp.ai.mit.edu:pub/alan/pclsr.memo).

; RESUME-PROC is called when the image is resumed.
; This does a garbage collection rooting from RESUME-PROC, writes the heap
; into a file, and then aborts the garbage collection (which didn't modify
; any VM registers or the stack).

(define-primitive write-image (string-> any-> string->)
  (lambda (filename resume-proc comment-string)
    (let ((lose (lambda (reason)
		  (raise-exception* reason 0
				    filename resume-proc comment-string))))
      (if (not (image-writing-okay?))
	  (lose (enum exception unimplemented-instruction))
	  (let ((fn (extract-string filename)))
	    (if (null-pointer? fn)
		(lose (enum exception out-of-memory))
		(let ((port (open-output-file fn)))
		  (deallocate fn)
		  (cond ((null-port? port)
			 (lose (enum exception cannot-open-channel)))
			((not (begin-simple-i/o))
			 (lose (enum exception out-of-memory)))
			(else
			 (let ((cs (extract-string comment-string)))
			   (cond ((null-pointer? cs)
				  (lose (enum exception out-of-memory)))
				 (else
				  (maybe-write-string cs port)
				  (deallocate cs)
				  (begin-collection)
				  (let ((resume-proc (trace-value resume-proc)))
				    (do-gc)
				    (let ((size (write-image port resume-proc)))
				      (abort-collection)
				      (if (and (finish-simple-i/o)
					       (close-output-port port))
					  (goto return (enter-fixnum size))
					  (lose (enum exception os-error)))))))))))))))))
