; -*- Mode: Scheme; Syntax: Scheme; Package: Scheme; -*-
; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.


; This is file prim.scm.
; Requires DEFINE-PRIMITIVE macro.

;;;; VM data manipulation primitives

; Input checking and coercion

(define (input-type pred coercer)  ;Alonzo wins
  (lambda (f) (f pred coercer)))

(define (input-type-predicate type) (type (lambda (x y) y x)))
(define (input-type-coercion type)  (type (lambda (x y) x y)))

(define (no-coercion x) x)

(define any->         (input-type (lambda (x) x #t) no-coercion))
(define fixnum->      (input-type fixnum?      extract-fixnum))
(define char->        (input-type vm-char?     extract-char))
(define vm-char->     (input-type vm-char?     no-coercion))
(define boolean->     (input-type vm-boolean?  extract-boolean))
(define location->    (input-type location?    no-coercion))
(define string->      (input-type vm-string?   no-coercion))
(define vector->      (input-type vm-vector?   no-coercion))
(define code-vector-> (input-type code-vector? no-coercion))

; Output coercion

(define (return val)
  (set! *val* val)
  (goto interpret))

(define return-any return)

(define (return-boolean x)
  (goto return (enter-boolean x)))

(define (return-fixnum x)
  (goto return (enter-fixnum x)))

(define (return-char x)
  (goto return (enter-char x)))

(define (return-unspecific x)
  x ;ignored
  (goto return unspecific))

; Scalar primitives

(define-primitive eq? (any-> any->) vm-eq? return-boolean)

; Rudimentary generic arithmetic.  Incomplete and confusing.
;  How to modularize for VM's like Maclisp that have generic arithmetic
;  built-in?

; These predicates are used to characterize the numeric representations that
; are implemented in the VM.

(define primitive-number?  fixnum?)
(define primitive-real?    fixnum?)
(define primitive-integer? fixnum?)

(define number->  (input-type primitive-number?  no-coercion))
(define real->    (input-type primitive-real?    no-coercion))
(define integer-> (input-type primitive-integer? no-coercion))

(define-primitive number? (any->)
  (lambda (x)
    (or (fixnum? x)
	(bignum? x)
	(ratnum? x)
	(double? x)
	(extended-number? x)))
  return-boolean)

(define-primitive integer?
  (any->)
  (lambda (n)
    (cond ((or (fixnum? n)
	       (bignum? n))
	   (goto return (enter-boolean #t)))
	  ((extended-number? n)
	   (raise-exception wrong-type-argument 0 n))
	  (else
	   (goto return (enter-boolean #f))))))

(define-syntax define-rational-predicate
  (syntax-rules ()
    ((define-numeric-predicate op)
     (define-primitive op
       (any->)
       (lambda (n)
	 (cond ((or (fixnum? n)
		    (bignum? n)
		    (ratnum? n)
		    (double? n))
		(goto return (enter-boolean #t)))
	       ((extended-number? n)
		(raise-exception wrong-type-argument 0 n))
	       (else
		(goto return (enter-boolean #f)))))))))

(define-rational-predicate rational?)
(define-rational-predicate real?)
(define-rational-predicate complex?)

; These primitives have a simple answer in the case of fixnums; for all other
; they punt to the run-time system.

(define-primitive exact?      (number->) (lambda (n) #t) return-boolean)
(define-primitive real-part   (number->) (lambda (n) (goto return n)))
(define-primitive imag-part   (number->) (lambda (n)
					      (goto return (enter-fixnum 0))))
(define-primitive floor       (number->) (lambda (n) (goto return n)))
(define-primitive numerator   (number->) (lambda (n) (goto return n)))
(define-primitive denominator (number->) (lambda (n)
					      (goto return (enter-fixnum 1))))
(define-primitive angle (number->)
  (lambda (n)
    (if (>= n 0)
	(goto return (enter-fixnum 0))
	(raise-exception arithmetic-overflow 0 n))))

; beware of (abs least-fixnum)
(define-primitive magnitude (number->)
  (lambda (n)
    (abs-carefully n
		   return
		   (lambda (n)
		     (raise-exception arithmetic-overflow 0 n)))))

; These primitives all just raise an exception and let the run-time system do
; the work.

(define-syntax define-punt-primitive
  (syntax-rules ()
    ((define-punt-primitive op)
     (define-primitive op (number->)
       (lambda (n) (raise-exception arithmetic-overflow 0 n))))))

(define-punt-primitive exact->inexact)
(define-punt-primitive inexact->exact)
(define-punt-primitive exp)
(define-punt-primitive log)
(define-punt-primitive sin)
(define-punt-primitive cos)
(define-punt-primitive tan)
(define-punt-primitive asin)
(define-punt-primitive acos)
(define-punt-primitive sqrt)

(define-syntax define-punt2-primitive
  (syntax-rules ()
    ((define-punt2-primitive op)
     (define-primitive op (number-> number->)
       (lambda (n m) (raise-exception arithmetic-overflow 0 n m))))))

(define-punt2-primitive atan)
(define-punt2-primitive make-polar)
(define-punt2-primitive make-rectangular)

(define (arithmetic-overflow x y)
  (raise-exception arithmetic-overflow 0 x y))

(define (arith op)
  (lambda (x y)
    (goto op x y return arithmetic-overflow)))

(define-primitive +         (number->  number->) (arith add-carefully))
(define-primitive -         (number->  number->) (arith subtract-carefully))
(define-primitive *         (number->  number->) (arith multiply-carefully))
(define-primitive /         (number->  number->) (arith divide-carefully))
(define-primitive quotient  (integer-> integer->) (arith quotient-carefully))
(define-primitive remainder (integer-> integer->) (arith remainder-carefully))
(define-primitive =         (number->  number->) vm-= return-boolean)
(define-primitive <         (real->    real->)   vm-< return-boolean)
(define-primitive arithmetic-shift (number-> number->)
  (arith shift-carefully))

(define-primitive char?       (any->) vm-char? return-boolean)
(define-primitive char=?      (vm-char-> vm-char->) vm-char=? return-boolean)
(define-primitive char<?      (vm-char-> vm-char->) vm-char<? return-boolean)
(define-primitive char->ascii (char->) char->ascii return-fixnum)

(define-primitive ascii->char
  (fixnum->)
  (lambda (x)
    (if (or (> x 255) (< x 0))
        (raise-exception wrong-type-argument 0 (enter-fixnum x))
	(goto return (enter-char (ascii->char x))))))

(define-primitive eof-object?
  (any->)
  (lambda (x) (vm-eq? x eof-object))
  return-boolean)

(define-primitive bitwise-not (fixnum->)          bitwise-not return-fixnum)
(define-primitive bitwise-and (fixnum-> fixnum->) bitwise-and return-fixnum)
(define-primitive bitwise-ior (fixnum-> fixnum->) bitwise-ior return-fixnum)
(define-primitive bitwise-xor (fixnum-> fixnum->) bitwise-xor return-fixnum)


(define-primitive stored-object-has-type?
  (any->)
  (lambda (x)
    (stob-of-type? x (next-byte)))
  return-boolean)

(define-primitive stored-object-length
  (any->)
  (lambda (stob)
    (let ((type (next-byte)))
      (if (stob-of-type? stob type)
	  (goto return-fixnum (d-vector-length stob))
	  (raise-exception wrong-type-argument 1 stob (enter-fixnum type))))))
  
; Fixed sized objects

(define-primitive stored-object-ref
  (any->)
  (lambda (stob)
    (let* ((type (next-byte))
	   (offset (next-byte)))
      (if (stob-of-type? stob type)
	  (goto return (d-vector-ref stob offset))
	  (raise-exception wrong-type-argument 2
			   stob
			   (enter-fixnum type)
			   (enter-fixnum offset))))))

(define-primitive stored-object-set!
  (any-> any->)
  (lambda (stob value)
    (let* ((type (next-byte))
	   (offset (next-byte)))
      (cond ((and (stob-of-type? stob type)
		  (not (immutable? stob)))
	     (d-vector-set! stob offset value)
	     (goto return unspecific))
	    (else
	     (raise-exception wrong-type-argument 2
			      stob
			      (enter-fixnum type)
			      (enter-fixnum offset)
			      value))))))

; Indexed objects

(define-primitive stored-object-indexed-ref
  (any-> fixnum->)
  (lambda (stob index)
    (let ((type (next-byte)))
      (cond ((not (stob-of-type? stob type))
	     (raise-exception wrong-type-argument 1
			      stob (enter-fixnum type) (enter-fixnum index)))
	    ((valid-index? index (d-vector-length stob))
	     (goto return (d-vector-ref stob index)))
	    (else
	     (raise-exception index-out-of-range 1
			      stob (enter-fixnum type) (enter-fixnum index)))))))

(define-primitive stored-object-indexed-set! (any-> fixnum-> any->)
  (lambda (stob index value)
    (let ((type (next-byte)))
      (cond ((or (not (stob-of-type? stob type))
		 (immutable? stob))
	     (raise-exception wrong-type-argument 1
			      stob (enter-fixnum type)
			      (enter-fixnum index) value))
	    ((valid-index? index (d-vector-length stob))
	     (d-vector-set! stob index value)
	     (goto return unspecific))
	    (else
	     (raise-exception index-out-of-range 1
			      stob (enter-fixnum type)
			      (enter-fixnum index) value))))))

; Hacko record handlers done for speed.

(define-primitive checked-record-ref
  (any-> any-> fixnum->)
  (lambda (record type index)
    (cond ((not (and (stob-of-type? record (enum stob record))
		     (vm-eq? type (record-ref record 0))))
	   (raise-exception wrong-type-argument 0
			    record type (enter-fixnum index)))
	  ((valid-index? index (record-length record))
	   (goto return (record-ref record index)))
	  (else
	   (raise-exception index-out-of-range 0
			    record type (enter-fixnum index))))))

(define-primitive checked-record-set! (any-> any-> fixnum-> any->)
  (lambda (record type index value)
    (cond ((not (and (stob-of-type? record (enum stob record))
		     (vm-eq? type (record-ref record 0))
		     (not (immutable? record))))
	   (raise-exception wrong-type-argument 0
			    record type (enter-fixnum index) value))
	  ((valid-index? index (record-length record))
	   (record-set! record index value)
	   (goto return unspecific))
	  (else
	   (raise-exception index-out-of-range 0
			    record type (enter-fixnum index) value)))))

; Constructors

(define-primitive make-stored-object ()
  (lambda ()
    (let* ((key (ensure-space (cells->bytes (+ stob-overhead (this-byte)))))
	   (len (next-byte))  ; can't consume this byte until after ENSURE-SPACE
	   (new (make-d-vector (next-byte) len key)))
      (cond ((>= len 1)
	     (d-vector-set! new (- len 1) *val*)
	     (do ((i (- len 2) (- i 1)))
		 ((> 0 i)
		  (unassigned))  ; for the type checker!
	       (d-vector-set! new i (pop)))))
      new))
  return)

; This is for the closed compiled versions of VECTOR and RECORD.

(define-primitive closed-make-stored-object ()
  (lambda ()
    (let* ((len *nargs*)
	   (key (ensure-space (cells->bytes (+ stob-overhead len))))
	   (new (make-d-vector (next-byte) len key)))
      (do ((i (- len 1) (- i 1)))
	  ((> 0 i)
	   (unassigned))  ; for the type checker!
	(d-vector-set! new i (pop)))
      new))
  return)

(define-primitive make-vector-object (fixnum-> any->)
  (lambda (len init)
    (let ((type (next-byte))
	  (size (vm-vector-size len)))
      (if (or (< len 0)
	      (> size max-stob-size-in-cells))
	  (raise-exception wrong-type-argument 1
			   (enter-fixnum type) (enter-fixnum len) init)
	  (maybe-ensure-space-saving-temp size init
            (lambda (okay? key init)
	      (if (not okay?)
		  (raise-exception heap-overflow 1
				   (enter-fixnum type) (enter-fixnum len) init)
		  (let ((v (make-d-vector type len key)))
		    (do ((i (- len 1) (- i 1)))
			((< i 0))
		      (d-vector-set! v i init))
		    (goto return v)))))))))

; Strings and code-vectors

(define-primitive string-length
  (string->)
  (lambda (string)
    (goto return-fixnum (vm-string-length string))))

(define-primitive code-vector-length
  (code-vector->)
  (lambda (code-vector)
    (goto return-fixnum (code-vector-length code-vector))))

(define (make-byte-ref ref length returner)
  (lambda (vector index)
    (if (valid-index? index (length vector))
	(goto returner (ref vector index))
	(raise-exception index-out-of-range 0 vector (enter-fixnum index)))))

(define-primitive string-ref (string-> fixnum->)
  (make-byte-ref vm-string-ref vm-string-length return-char))

(define-primitive code-vector-ref (code-vector-> fixnum->)
  (make-byte-ref code-vector-ref code-vector-length return-fixnum))

(define (make-byte-setter setter length enter-elt)
  (lambda (vector index char)
    (cond ((immutable? vector)
	   (raise-exception wrong-type-argument 1
			    vector (enter-fixnum index) (enter-elt char)))
	  ((valid-index? index (length vector))
	   (setter vector index char)
	   (goto return unspecific))
	  (else
	   (raise-exception index-out-of-range 1
			    vector (enter-fixnum index) (enter-elt char))))))

(define-primitive string-set! (string-> fixnum-> char->)
  (make-byte-setter vm-string-set! vm-string-length enter-char))

(define-primitive code-vector-set! (code-vector-> fixnum-> fixnum->)
  (make-byte-setter code-vector-set! code-vector-length enter-fixnum))

(define (byte-vector-maker size maker setter enter-elt)
  (lambda (len init)
    (let ((size (size len)))
      (if (or (< len 0)
	      (> size max-stob-size-in-cells))
	  (raise-exception wrong-type-argument 0
			   (enter-fixnum len) (enter-elt init))
	  (maybe-ensure-space-saving-temp
	   size
	   (enter-fixnum 0)
	   (lambda (okay? key ignore)
	     (if (not okay?)
		 (raise-exception heap-overflow 0
				  (enter-fixnum len) (enter-elt init))
		 (let ((vector (maker len key)))
		   (do ((i (- len 1) (- i 1)))
		       ((< i 0))
		     (setter vector i init))
		   (goto return vector)))))))))

(define-primitive make-string (fixnum-> char->)
  (byte-vector-maker vm-string-size vm-make-string vm-string-set! enter-char))

(define-primitive make-code-vector (fixnum-> fixnum->)
  (byte-vector-maker code-vector-size make-code-vector code-vector-set!
		     enter-fixnum))

; Locations & mutability

(define-primitive location-defined? (location->)
  (lambda (loc)
    (return-boolean (or (not (undefined? (contents loc)))
			(= (contents loc) unassigned-marker)))))

(define-primitive set-location-defined?! (location-> boolean->)
  (lambda (loc value)
    (cond ((not value)
	   (set-contents! loc unbound-marker))
	  ((undefined? (contents loc))
	   (set-contents! loc unassigned-marker)))
    (goto return unspecific)))

(define-primitive immutable? (any->) immutable? return-boolean)

(define-primitive make-immutable! (any->)
  (lambda (thing)
    (make-immutable! thing)
    (goto return thing)))


; Misc

(define-primitive false ()
  (lambda ()
    (goto return false)))

(define-primitive eof-object ()
  (lambda ()
    (goto return eof-object)))

(define-primitive trap (any->)
  (lambda (arg)
    (raise-exception trap 0 arg)))

(define-primitive find-all-symbols (vector->)
  (lambda (table)
    (let ((do-it (lambda (table)
		   (walk-over-symbols
		    (lambda (symbol)
		      (if (available? vm-pair-size)
			  (let ((key (preallocate-space vm-pair-size)))
			    (add-to-symbol-table symbol table key)
			    #t)
			  #f))))))
      (if (or (do-it table)
	      (do-it (collect-saving-temp table)))
	  (goto return unspecific)
	  (raise-exception heap-overflow 0)))))

(define-primitive find-all-xs (fixnum->)
  (lambda (type)
    (let ((vector (find-all-xs type)))
      (cond ((not (false? vector))
	     (goto return vector))
	    (else
	     (collect)
	     (let ((vector (find-all-xs type)))
	       (if (not (false? vector))
		   (goto return vector)
		   (raise-exception heap-overflow 0 (enter-fixnum type)))))))))

(define-primitive collect ()
  (lambda ()
    (set! *val* unspecific)
    (collect)
    (goto return unspecific)))

(define-primitive memory-status (fixnum-> any->)
  (lambda (key other)
    (cond ((= key (enum memory-status-option pointer-hash))
	   (goto return (descriptor->fixnum other)))
	  ((= key (enum memory-status-option available))
	   (goto return (enter-fixnum (available))))
	  ((= key (enum memory-status-option heap-size))
	   (goto return
		 (enter-fixnum
		  (bytes->cells (heap-size)))))
	  ((= key (enum memory-status-option stack-size))
	   (goto return (enter-fixnum (stack-size))))
	  ((= key (enum memory-status-option gc-count))
	   (goto return (enter-fixnum (gc-count))))
	  ((= key (enum memory-status-option expand-heap!))
	   (raise-exception unimplemented-instruction 0
			    (enter-fixnum key) other))
	  (else
	   (raise-exception bad-option 0 (enter-fixnum key) other)))))

(define-primitive time (fixnum-> any->)
  (lambda (option other)
    (cond ((= option (enum time-option tick-mantissa))
	   (goto return (enter-fixnum (ps-clock-tick))))
	  ((= option (enum time-option tick-exponent))
	   (goto return (enter-fixnum (ps-clock-exponent))))
	  ((= option (enum time-option run-time))
	   (goto return (enter-fixnum (ps-run-time))))
	  ((= option (enum time-option real-time))
	   (goto return (enter-fixnum (ps-real-time))))
	  (else
	   (raise-exception bad-option 0 (enter-fixnum option) other)))))

(define-primitive schedule-interrupt (fixnum-> fixnum-> boolean->)
  (lambda (mantissa exponent periodic)
    (let ((time (adjust-time mantissa exponent)))
      (clear-interrupt! (enum interrupt alarm))
      (goto return (enter-fixnum (ps-schedule-interrupt time periodic))))))

; Convert from the user's exponent to the system's.

(define (adjust-time mantissa exponent)
  (let ((system (ps-clock-exponent)))
    (cond ((= exponent system)
	   mantissa)
	  ((> system exponent)
	   (quotient mantissa (expt 10 (- system exponent))))
	  (else
	   (* mantissa (expt 10 (- exponent system)))))))

(define external-> (input-type external? no-coercion))

(define-primitive external-lookup (external->)
  (lambda (external)
    (let ((name (external-name external))
	  (value (external-value external)))
      (if (and (vm-string? name)
	       (code-vector? value)
	       (lookup-external-name (address-after-header name)
				     (address-after-header value)))
	  (goto return unspecific)
	  (raise-exception unbound-external-name 0 external)))))

; This is only for the closed-compiled version of EXTERNAL-CALL.
;
; This is a mess because the arguments have been pushed on the stack after
; the procedure.
; The top-level driver loop removes the procedure and the arguments from
; the stack.

(define-primitive external-call ()
  (lambda ()
    (let ((proc (stack-ref (- *nargs* 1))))
      (cond ((not (external? proc))    ; lots of junk on the stack...
	     (raise-exception wrong-type-argument 0 proc))
	    (else
	     (set! *nargs* (- *nargs* 1))      ; don't count the procedure
	     (set! *val* proc)
	     (enum return-option external-call))))))  ; return to driver loop

(define-enumeration vm-extension-status
  (okay
   exception
   ))

(define *extension-value* (unassigned))

(define-primitive vm-extension (fixnum-> any->)
  (lambda (key value)
    (let ((status (extended-vm key value)))
      (cond ((vm-eq? status (enum vm-extension-status okay))
	     (goto return *extension-value*))
	    ((vm-eq? status (enum vm-extension-status exception))
	     (raise-exception extension-exception 0 (enter-fixnum key) value))
	    (else
	     (raise-exception extension-return-error 0
			      (enter-fixnum key) value))))))

(define-primitive vm-return (fixnum-> any->)
  (lambda (key value)
    (set! *val* value)
    (enum return-option exit)))            ; the VM returns this value

(define-primitive current-thread ()
  (lambda () *current-thread*)
  return-any)

(define-primitive set-current-thread! (any->)
  (lambda (state)
    (set! *current-thread* state)
    unspecific)
  return-any)

; Unnecessary primitives

(define-primitive string=? (string-> string->) vm-string=? return-boolean)

; Special primitive called by the reader.
; Primitive for the sake of speed.  Probably should be flushed.

(define-consing-primitive reverse-list->string (any-> fixnum->) 
  (lambda (n) (vm-string-size (extract-fixnum n)))
  (lambda (l n k)
    (if (not (or (vm-pair? l) (vm-eq? l null)))
        (raise-exception wrong-type-argument 0 l (enter-fixnum n))
        (let ((obj (vm-make-string n k)))
          (do ((l l (vm-cdr l))
               (i (- n 1) (- i 1)))
              ((< i 0) (goto return obj))
            (vm-string-set! obj i (extract-char (vm-car l))))))))

(define-primitive string-hash (string->) vm-string-hash return-fixnum)

(define-consing-primitive intern (string-> vector->)
  (lambda (ignore) (+ vm-symbol-size vm-pair-size))
  intern
  return)

; Messy because we have to detect circular lists (alternatively we
; could check for interrupts and then pclsr (see above)).

(define-primitive assq (any-> any->)
  (lambda (thing list)
    (let ((lose (lambda ()
		  (raise-exception wrong-type-argument 0 thing list))))
      (let loop ((list list) (slow list) (move-slow? #t))
	(cond ((vm-eq? list null)
	       (enter-boolean #f))
	      ((not (vm-pair? list))
	       (lose))
	      (else
	       (let ((head (vm-car list)))
		 (cond ((not (vm-pair? head))
			(lose))
		       ((vm-eq? (vm-car head) thing)
			head)
		       (else
			(let ((list (vm-cdr list)))
			  (cond ((eq? list slow)
				 (lose))
				(move-slow?
				 (loop list (vm-cdr slow) #f))
				(else
				 (loop list slow #t))))))))))))
  return)

; Eventually add make-table, table-ref, table-set! as primitives?
; No -- write a compiler instead.
