; -*- Mode: Scheme; Syntax: Scheme; Package: Scheme; -*-
; Copyright (c) 1993, 1994 Richard Kelsey and Jonathan Rees.  See file COPYING.


; This is file vmio.scm.

; Questions:
;  What to do if an error occurs?
;  How to deal with transcript-on and transcript-off ?
;  How to deal with uninterrubtibly opening a port and pushing it onto
;    an outstanding-ports list?

; *open-vm-channels* is a vector of descriptors for open channels.
; *open-ports* is a vector of open ports of the underlying Scheme
; implementation.

(define *number-of-channels* 100)

(define *open-ports* (unassigned))
(define *open-vm-channels* (unassigned))

; Channel modes

(define for-input      1)
(define for-output     2)
(define special-input  3)  ; socket accept ports, etc.
(define special-output 4)  ; ???

(define (input-channel? channel)
  (= (channel-mode channel) (enter-fixnum for-input)))

(define (output-channel? channel)
  (= (channel-mode channel) (enter-fixnum for-output)))

(define iic-index 0)  ;index of initial input channel
(define ioc-index 1)  ;[out/in]ditto
(define iec-index 2)  ;[error/in]ditto

(define (initialize-i/o-system)
  (set! *open-ports* (make-vector *number-of-channels*))
  (set! *open-vm-channels* (make-vector *number-of-channels*))
  (if (or (null-pointer? *open-ports*)
	  (null-pointer? *open-vm-channels*))
      (error "out of memory, unable to continue"))
  (vector+length-fill! *open-vm-channels* *number-of-channels* false)
  (make-initial-channel iic-index for-input  (current-input-port)  "si")
  (make-initial-channel ioc-index for-output (current-output-port) "so")
  (make-initial-channel iec-index for-output (current-error-port)  "se")
  (set! *waiting-for-input* (make-port-set))
  (set! *waiting-for-output* (make-port-set))
  (set! *waiting-for-trouble* (make-port-set))
  unspecific)

(define (make-initial-channel index mode port name)
  (vector-set! *open-ports* index port)
  (vector-set! *open-vm-channels*
	       index
	       (make-channel (enter-fixnum mode)
			     (enter-fixnum index)
			     (enter-string name)
			     universal-key)))

(define (allocate-another-channel-descriptor got-one-cont none-left-cont)
  (let* ((new-count (+ *number-of-channels* 8))
	 (old-count *number-of-channels*)
	 (new-ports (make-vector new-count))
	 (new-vm-channels (make-vector new-count)))
    (cond ((or (null-pointer? new-ports)
	       (null-pointer? new-vm-channels))
	   (none-left-cont))
	  (else
	   (vector+length-fill! new-vm-channels new-count false)
	   (do ((i 0 (+ i 1)))
	       ((>= i *number-of-channels*))
	     (vector-set! new-ports i (vector-ref *open-ports* i))
	     (vector-set! new-vm-channels i (vector-ref *open-vm-channels* i)))
	   (deallocate *open-ports*)
	   (deallocate *open-vm-channels*)
	   (set! *open-ports* new-ports)
	   (set! *open-vm-channels* new-vm-channels)
	   (set! *number-of-channels* new-count)
	   (got-one-cont old-count)))))  ; new and unallocated channel index

; A channel, port, and buffer for each of initial-input-port, initial-output-
; port, and initial-error-port.  The error port has a buffer of size zero.
; Each channel has a two character name.

(define initial-i/o-heap-space
  (+ (* 3 (+ (+ channel-size (vm-string-size 2))
	     port-size))
     (+ (* 2 (code-vector-size 1024))
	(code-vector-size 0))))

(define (initial-input-channel)
  (vector-ref *open-vm-channels* iic-index))

(define (initial-output-channel)
  (vector-ref *open-vm-channels* ioc-index))

(define (initial-error-channel)
  (vector-ref *open-vm-channels* iec-index))

; Auxiliaries for I/O primitives

(define (extract-channel channel)
  (vector-ref *open-ports* (extract-fixnum (channel-index channel))))

(define (find-channel-index)
  (let loop ((i 0))
    (cond ((>= i *number-of-channels*)
	   -1)
	  ((false? (vector-ref *open-vm-channels* i))
	   i)
	  (else (loop (+ i 1))))))

(define (use-channel-index! index vm-port port)
  (vector-set! *open-ports* index port)
  (vector-set! *open-vm-channels* index vm-port))

(define (use-raw-channel-index! index vm-port raw-port)
  (vector-set! *open-ports* index raw-port)
  (vector-set! *open-vm-channels* index vm-port))

; [An open can fail for several reasons:
;    - The OS couldn't open it
;    - There are no empty slots in *open-ports* and we can't get more memory
; ]

(define (open-channel spec mode key collect-saving-temp
		      succeeded os-failed)
  (let loop ((index (find-channel-index)) (spec spec))
    (if (>= index 0)
	(really-open-channel spec index mode key succeeded os-failed)
	(let ((spec (collect-saving-temp spec)))
	  (let ((index (find-channel-index)))
	    (if (>= index 0)
		(loop index spec)
		(allocate-another-channel-descriptor
		 (lambda (index)
		   (loop index spec))
		 (lambda ()
		   (os-failed spec (enum exception out-of-memory))))))))))

(define (really-open-channel spec index mode key succeeded os-failed)
  (if (or (= mode (enum open-channel-option input-file))
	  (= mode (enum open-channel-option output-file)))
      (let ((filename (extract-string spec)))
	(if (null-pointer? filename)
	    (os-failed spec (enum exception out-of-memory))
	    (let ((port (if (= mode (enum open-channel-option input-file))
			    (open-input-file filename)
			    (open-output-file filename))))
	      (deallocate filename)
	      (if (null-port? port)
		  (os-failed spec (enum exception cannot-open-channel))
		  (succeeded (actually-make-channel mode index spec key port))))))
      ;; fixnum specs work for Unix, what to do in general?
      ;; Coercion is alway ugly.
      (succeeded
       (actually-make-channel mode index spec key
			      (coerce (extract-fixnum spec)
				      'int32
				      'port)))))
			      

(define (actually-make-channel mode index spec key port)
  (let ((channel (make-channel (enter-fixnum
				(translate-channel-mode mode))
			       (enter-fixnum index)
			       spec
			       key)))
    (use-channel-index! index channel port)
    channel))

(define (translate-channel-mode mode)
  (cond ((or (= mode (enum open-channel-option input-file))
	     (= mode (enum open-channel-option raw-input-channel)))
	 for-input)
	((or (= mode (enum open-channel-option output-file))
	     (= mode (enum open-channel-option raw-output-channel)))
	 for-output)
	((= mode (enum open-channel-option special-input-channel))
	 special-input)
	(else ; (= mode (enum open-channel-option special-input-channel))
	 special-output)))

(define (open? channel)
  (>= (channel-index channel) (enter-fixnum 0)))

(define (close-channel channel)
  (if (open? channel)
      (let ((index (extract-fixnum (channel-index channel))))
        (if (not (or (= index iic-index)
                     (= index ioc-index)
                     (= index iec-index)))
            (let ((port (extract-channel channel)))
              (set-channel-index! channel (enter-fixnum -1))
              (vector-set! *open-vm-channels*  index false)
	      (if (or (= for-input (channel-mode channel))
		      (= special-input (channel-mode channel)))
		  (close-input-port port)
		  (close-output-port port)))
	    #f))
      #t))

; The following is called after the GC finishes.

(define (close-untraced-channels!)
  (do ((i 0 (+ i 1)))
      ((= i *number-of-channels*) #f)
    (let ((channel (vector-ref *open-vm-channels* i)))
      (if (not (false? channel))
	  (let* ((header (stob-header channel))
		 (new (cond ((stob? header)  ; channel was copied
			     header)
			    (else            ; channel was not copied
			     (close-channel-noisily channel)
			     false))))
	    (vector-set! *open-vm-channels* i new))))))

(define (close-channel-noisily channel)
  (close-channel channel)
  (write-string "Channel closed: " (current-error-port))
  (if (fixnum? (channel-id channel))
      (write-string "not a file channel" (current-error-port))
      (let* ((header (stob-header (channel-id channel)))
	     (name (if (stob? header)
		       header              ; name was copied
		       (channel-id channel))))   ; name was not copied
	(write-vm-string name (current-error-port))))
  (newline (current-error-port)))

(define (write-vm-string string port)
  (write-block port
	       (address-after-header string)
	       (vm-string-length string)
	       (lambda (okay? count)
		 (if (not (and okay? (= count (vm-string-length string))))
		     (error "cannot write to error output")))))

; Finding ports that are ready for use.

(define *waiting-for-input*  (unassigned))
(define *waiting-for-output* (unassigned))
(define *waiting-for-trouble* (unassigned))

(define (find-ready-channels channels wait-time key)
  (let ((input-set *waiting-for-input*)
	(output-set *waiting-for-output*)
	(trouble-set *waiting-for-trouble*))
    (clear-port-set! input-set)
    (clear-port-set! output-set)
    (clear-port-set! trouble-set)
    (do ((channels channels (vm-cdr channels)))
	((vm-eq? null channels))
      (let* ((channel (vm-car channels))
	     (mode (channel-mode channel)))
	(add-to-port-set! (if (or (= mode (enter-fixnum for-input))
				  (= mode (enter-fixnum special-input)))
			      input-set
			      output-set)
			  (extract-channel channel))
	(add-to-port-set! trouble-set (extract-channel channel))))
    (case (find-ready-ports input-set output-set trouble-set wait-time)
      ((0) null)    ; ran out of time OR an interrupt occured
      ((-1) false)  ; an error occurred
      (else
       (select-ready-channels channels input-set output-set trouble-set key)))))

; The order of the channels in the returned list is the same as that
; in CHANNELS.

(define (select-ready-channels channels input-set output-set trouble-set key)
  (let loop ((channels channels) (ready null) (last null))
    (if (vm-eq? null channels)
	ready
	(let* ((channel (vm-car channels))
	       (channels (vm-cdr channels))
	       (mode (channel-mode channel)))
	  (if (or (port-set-member? (if (or (= mode (enter-fixnum for-input))
					    (= mode (enter-fixnum special-input)))
					input-set
					output-set)
				    (extract-channel channel))
		  (port-set-member? trouble-set (extract-channel channel)))
	      (let ((p (vm-cons channel null key)))
		(cond ((vm-eq? ready null)
		       (loop channels p p))
		      (else
		       (vm-set-cdr! last p)
		       (loop channels ready p))))
	      (loop channels ready last))))))

